var package = {
		hashlib: require('./modules/hashlib/hashlib'),
		Redis: require('redis'),
		DbConn: require('./modules/DbConn'),
		ClearSilver: require('ClearSilver'),
		auditor: require('auditor'),
		Kahana: require('Kahana')
	},
	status = {
		OK: 'OK',
		SERVER_ERROR: 'SERVER_ERROR',
		INVALID_REQUEST: 'INVALID_REQUEST',
		NO_ACCOUNT: 'NO_ACCOUNT',
		NO_UNIQUE: 'NO_UNIQUE',
		NO_TICKET: 'NO_TICKET'
	};

function Inspect( obj, level ){
	console.log( require('util').inspect( obj, true, level ) );
}

// setup validation rule
var auditor = new package.auditor();
SetRules();

Application.prototype.SetRules = function()
{
	var rules = {
			TicketIssue: {
				email: {
					type: 'email',
					required: true,
					min: 6
				}
			},
			Signup: {
				ticket: {
					type: 'text',
					required: true,
					min: 40,
					max: 40
				},
				agree: {
					type: 'unsigned',
					required: true,
					fix: 1
				},
				uname: {
					type: 'text',
					required: true,
					min: 2,
					max: 30
				},
				pswd: {
					type: 'text',
					required: true,
					min: 6,
					max: 20
				},
				birth: {
					type: 'text',
					min: 8,
					max: 10
				},
				gender: {
					type: 'unsigned',
					min: 0,
					max: 2
				}
			},
			Reset: {
				ticket: {
					type: 'text',
					required: true,
					min: 40,
					max: 40
				},
				pswd: {
					type: 'text',
					required: true,
					min: 6,
					max: '20'
				}
			},
			Signin: {
				email: {
					type: 'email',
					required: 1
				},
				pswd: {
					type: 'text',
					required: true
				}
			}
		};
	for( var label in rules ){
		this.auditor.add( label, rules[label] );
	}
};

// MARK: database connection
Application.prototype.Connect = function()
{
	// create database connection
	this.dbconn.Connect( function( err )
	{
		if( err ){
			console.log( 'error' );
			console.log( err );
		}
	},
	{
		host: '127.0.0.1',
		port: 5432,
		user: 'kontext',
		password: 'kontext0x211.pgsql!',
		database: 'kontext'
	} );
};

// MARK: DelegateRendering On
Application.prototype.SetDocumentRoot = function( docroot ){
	// nothing to do
	this.clearsilver.SetDocumentRoot( docroot );
};
Application.prototype.RenderString = function( str, obj ){
	return this.clearsilver.RenderString( str, obj );
};

// Db: delegate methods
// template rendering error
Application.prototype.ParseError = function( err ){
	console.log( err.stack );
};

// MARK: Required methods
Application.prototype.DeadEnd = function( last, r )
{
	if( r.status ){
		this.server.RequestClose( r, status );
	}
	else {
		var self = this;
		this.server.MapToStorage( r, function( err, status, r ){
			self.server.RequestClose( r, ( status ) ? status : 200 );
		} );
	}
};

Application.prototype.Detour = function( route, r, method, args, last, runNext )
{
	// this.AddLog( r, 'route[' + route + '][Detour]: call -> ' + method + ', args -> ' + args + ', last -> ' + last + ', runNext: ' + typeof( runNext ) );
	// console.log( 'route[' + route + '][Detour]: call -> ' + method + ', args -> ' + args + ', last -> ' + last + ', runNext: ' + typeof( runNext ) );
	this.DeadEnd( true, r );
};


// MARK: ROUTE
Application.prototype.RequestCheck = function( r, rule, callback )
{
	if( r.req.method == 'POST' )
	{
		var self = this,
			finish = function( err, r )
			{
				if( err ){
					r.hdf.App.Error = err;
					self.DeadEnd( true, r );
				}
				else
				{
					var fields = self.auditor.getFields( rule ),
						audit = { errno: 0, items:{} },
						field,item;
					
					while( fields.length )
					{
						field = fields.shift();
						if( ( item = self.auditor.check( rule, field, r.data.fields[field] ) ) )
						{
							if( !audit.errno ){
								audit.errno = item.errno;
							}
							audit.items[field] = item;
						}
					}
					
					r.hdf.App = {
						Audit: audit,
						Form: audit.items
					};
					callback.apply( self, [r] );
				}
			};
		
		r.hdf.App = {};
		// get request data
		if( this.server.RequestData( r, undefined, finish ) ){
			// close request if return not 0
			this.server.RequestClose( r );
		}
	}
	else {
		this.DeadEnd( true, r );
	}
};
// MARK: Ticket Handler
Application.prototype.Ticket = function( route, r, method, args, last, runNext )
{
	console.log( 'route[' + route + ']: call -> ' + method + ', args -> ' + args + ', last -> ' + last + ', runNext: ' + typeof( runNext ) );
	
	this.RequestCheck( r, 'TicketIssue', function( r )
	{
		Inspect( r.hdf, 6 );
		
		// default internal server error
		r.hdf.App.Status = status.SERVER_ERROR;
		if( r.hdf.App.Audit.errno !== 0 ){
			r.hdf.App.Status = status.INVALID_REQUEST;
			this.DeadEnd( true, r );
		}
		else
		{
			var self = this,
				commit = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						ctx.hdf.App.Status = status.SERVER_ERROR;
					}
					this.DeadEnd( true, ctx );
				},
				rollback = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						ctx.hdf.App.Status = status.SERVER_ERROR;
					}
					this.DeadEnd( true, ctx );
				},
				finalize = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else
					{
						console.log( result );
						ctx.hdf.App.Status = status.OK;
						delete ctx.hdf.App.Audit;
						delete ctx.hdf.App.Form;
						this.dbconn.Commit( commit, ctx );
					}
				},
				insert_ticket = function( err, result, ctx )
				{
					
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else if( result.rowCount ){
						ctx.hdf.App.Status = status.NO_UNIQUE;
						delete ctx.hdf.App.Audit;
						delete ctx.hdf.App.Form;
						this.dbconn.Commit( commit, ctx );
					}
					else
					{
						var email = ctx.hdf.App.Form.email.val,
							ticket = package.hashlib.hmac_sha1( email, String( Date.now() ) );
						
						this.dbconn.Query( 'INSERT INTO ticket ( email, ticket ) VALUES( $1, $2 ) RETURNING ticket;', 
											[email,ticket], finalize, ctx );
					}
				},
				select_email = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else {
						this.dbconn.Query( 'SELECT id FROM account WHERE email = $1', 
											[ctx.hdf.App.Form.email.val], insert_ticket, ctx );
					}
				},
				delete_ticket = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else {
						this.dbconn.Query( 'DELETE FROM ticket WHERE email = $1', 
											[ctx.hdf.App.Form.email.val], select_email, ctx );
					}
				};
			
			this.dbconn.Begin( delete_ticket, r );
		}
	} );
};

// MARK: Activate Handler
Application.prototype.Signup = function( route, r, method, args, last, runNext )
{
	console.log( 'route[' + route + ']: call -> ' + method + ', args -> ' + args + ', last -> ' + last + ', runNext: ' + typeof( runNext ) );
	// check ticket
	this.RequestCheck( r, 'Signup', function( r )
	{
		r.hdf.App.Status = status.SERVER_ERROR;
		// Inspect( audit );
		
		if( r.hdf.App.Audit.errno !== 0 ){
			r.hdf.App.Status = status.INVALID_REQUEST;
			this.DeadEnd( true, r );
		}
		else
		{
			var commit = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
					}
					else {
						ctx.hdf.App.Status = status.OK;
						delete ctx.hdf.App.Audit;
						delete ctx.hdf.App.Form;
					}
					this.DeadEnd( true, ctx );
				},
				rollback = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
					}
					this.DeadEnd( true, ctx );
				},
				finalize = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else if( !result.rowCount ){
						ctx.hdf.App.Status = status.NO_TICKET;
						this.dbconn.Commit( commit, ctx );
					}
					else {
						this.dbconn.Commit( commit, ctx );
					}
				},
				insert_profile = function( err, result, ctx )
				{
					// unknown ticket
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else if( !result.rowCount ){
						ctx.hdf.App.Status = status.NO_TICKET;
						this.dbconn.Rollback( rollback, ctx );
					}
					else
					{
						var row = result.rows[0],
							fields = ['email','clue','pswd','uname','birth','gender'],
							vals = [row.email];
						
						for( var i = 1; i < fields.length; i++ ){
							vals.push( ( ctx.hdf.App.Form[fields[i]] ) ? ctx.hdf.App.Form[fields[i]].val : null );
						}
						vals[1] = package.hashlib.hmac_sha1( vals[0] + vals[2], String( Date.now() ) );
						vals[2] = package.hashlib.hmac_sha1( vals[0] + vals[2], vals[1] );
						fields.push( 'agree' );
						vals.push( 'true' );
						this.dbconn.Query( 'INSERT INTO account ( ' + fields.join(',') + ' ) VALUES($1,$2,$3,$4,$5,$6,$7) RETURNING clue;', 
											vals, finalize, ctx );
					}
				},
				delete_ticket = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else {
						this.dbconn.Query( 'DELETE FROM ticket WHERE ticket = $1 RETURNING email;', [ctx.hdf.App.Form.ticket.val], insert_profile, ctx );
					}
				};
			
			this.dbconn.Begin( delete_ticket, r );
		}
	} );
};

// MARK: Password Recovery Handler
Application.prototype.Recovery = function( route, r, method, args, last, runNext )
{
	console.log( 'route[' + route + ']: call -> ' + method + ', args -> ' + args + ', last -> ' + last + ', runNext: ' + typeof( runNext ) );
	
	this.RequestCheck( r, 'TicketIssue', function( r )
	{
		// Inspect( audit );
		r.hdf.App.Status = status.SERVER_ERROR;
		if( r.hdf.App.Audit.errno !== 0 ){
			r.hdf.App.Status = status.INVALID_REQUEST;
			this.DeadEnd( true, r );
		}
		else
		{
			var self = this,
				commit = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						ctx.hdf.App.Status = status.SERVER_ERROR;
					}
					this.DeadEnd( true, ctx );
				},
				rollback = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						ctx.hdf.App.Status = status.SERVER_ERROR;
					}
					this.DeadEnd( true, ctx );
				},
				finalize = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else
					{
						console.log( result );
						ctx.hdf.App.Status = status.OK;
						delete ctx.hdf.App.Audit;
						delete ctx.hdf.App.Form;
						this.dbconn.Commit( commit, ctx );
					}
				},
				insert_ticket = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else if( !result.rowCount ){
						ctx.hdf.App.Status = status.NO_ACCOUNT;
						delete ctx.hdf.App.Audit;
						delete ctx.hdf.App.Form;
						this.dbconn.Commit( commit, ctx );
					}
					else
					{
						var row = result.rows[0],
							ticket = package.hashlib.hmac_sha1( row.email, String( Date.now() ) );
						
						this.dbconn.Query( 'INSERT INTO recovery ( email, ticket ) VALUES( $1, $2 ) RETURNING ticket;', 
											[row.email,ticket], finalize, ctx );
					}
				},
				select_email = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else {
						this.dbconn.Query( 'SELECT email FROM account WHERE email = $1', 
											[ctx.hdf.App.Form.email.val], insert_ticket, ctx );
					}
				},
				delete_ticket = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else {
						this.dbconn.Query( 'DELETE FROM recovery WHERE email = $1', 
											[ctx.hdf.App.Form.email.val], select_email, ctx );
					}
				};
			
			this.dbconn.Begin( delete_ticket, r );
		}
	} );
};

// MARK: Password Reset Handler
Application.prototype.Reset = function( route, r, method, args, last, runNext )
{
	console.log( 'route[' + route + ']: call -> ' + method + ', args -> ' + args + ', last -> ' + last + ', runNext: ' + typeof( runNext ) );
	
	this.RequestCheck( r, 'Reset', function( r )
	{
		// default internal server error
		r.hdf.App.Status = status.SERVER_ERROR;
		if( r.hdf.App.Audit.errno !== 0 ){
			r.hdf.App.Status = status.INVALID_REQUEST;
			this.DeadEnd( true, r );
		}
		else
		{
			var self = this,
				commit = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						ctx.hdf.App.Status = status.SERVER_ERROR;
					}
					this.DeadEnd( true, ctx );
				},
				rollback = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						ctx.hdf.App.Status = status.SERVER_ERROR;
					}
					this.DeadEnd( true, ctx );
				},
				finalize = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else
					{
						console.log( result );
						ctx.hdf.App.Status = status.OK;
						delete ctx.hdf.App.Audit;
						delete ctx.hdf.App.Form;
						this.dbconn.Commit( commit, ctx );
					}
				},
				reset_pswd = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					// unknown ticket
					else if( !result.rowCount ){
						ctx.hdf.App.Status = status.NO_TICKET;
						this.dbconn.Rollback( rollback, ctx );
					}
					else
					{
						var row = result.rows[0],
							pswd = ctx.hdf.App.Form.pswd.val,
							clue = package.hashlib.hmac_sha1( row.email + pswd, String( Date.now() ) );
						
						pswd = package.hashlib.hmac_sha1( row.email + pswd, clue );
						this.dbconn.Query( 'UPDATE account SET( pswd, clue ) = ( $1, $2 ) WHERE email = $3 RETURNING clue;', 
											[pswd,clue,row.email], finalize, ctx );
					}
				},
				delete_ticket = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.dbconn.Rollback( rollback, ctx );
					}
					else {
						this.dbconn.Query( 'DELETE FROM recovery WHERE ticket = $1 RETURNING email', 
											[ctx.hdf.App.Form.ticket.val], reset_pswd, ctx );
					}
				};
			
			this.dbconn.Begin( delete_ticket, r );
		}
	} );
};


// MARK: Login Handler
Application.prototype.Signin = function( route, r, method, args, last, runNext )
{
	console.log( 'route[' + route + ']: call -> ' + method + ', args -> ' + args + ', last -> ' + last + ', runNext: ' + typeof( runNext ) );
	// check ticket
	this.RequestCheck( r, 'Signin', function( r )
	{
		r.hdf.App.Status = status.SERVER_ERROR;
		if( r.hdf.App.Audit.errno !== 0 ){
			r.hdf.App.Status = status.INVALID_REQUEST;
			this.DeadEnd( true, r );
		}
		else
		{
			var self = this,
				// create session-id
				create_session = function( row, ctx )
				{
					var multi = this.redis.multi(),
						timestamp = new Date(),
						// session-cookie = hmac_sha1( id + uname + timestamp, mac + uname ) + timestamp + id
						//				  = hmac:40 + timestamp:13 + id:length
						session = package.hashlib.hmac_sha1( row.id + row.uname + +timestamp, row.clue );
					
					delete row.pswd;
					delete row.clue;
					multi.hmset( session, row );
					multi.expire( session, 86400 );
					multi.exec( function( err, result )
					{
						// login ok
						if( err ){
							ctx.hdf.App.Error = err;
							self.DeadEnd( true, ctx );
						}
						else {
							ctx.hdf.App.Status = status.OK;
							ctx.hdf.App.Session = session;
							delete ctx.hdf.App.Audit;
							delete ctx.hdf.App.Form;
							self.DeadEnd( true, ctx );
						}
					} );
				},
				check_user = function( err, result, ctx )
				{
					if( err ){
						ctx.hdf.App.Error = err;
						this.DeadEnd( true, ctx );
					}
					// unknown user
					else if( !result.rowCount ){
						ctx.hdf.App.Status = status.NO_ACCOUNT;
						this.DeadEnd( true, ctx );
					}
					else
					{
						var row = result.rows[0],
							val = package.hashlib.hmac_sha1( ctx.hdf.App.Form.email.val + ctx.hdf.App.Form.pswd.val, row.clue );
						
						// invalid password
						if( val !== row.pswd ){
							ctx.hdf.App.Status = status.NO_ACCOUNT;
							this.DeadEnd( true, ctx );
						}
						else {
							create_session.apply( this, [row,ctx] );
						}
					}
				};
			
			// check existing account by email
			this.dbconn.Query( 'SELECT id, uname, birth, gender, pswd, clue FROM account WHERE email = $1', [r.hdf.App.Form.email.val], check_user, r );
		}
	} );
};


// MARK: Signout Handler
Application.prototype.Signout = function( route, r, method, args, last, runNext )
{
	var expire = new Date(),
		domain = 'dev.src.kontext.jp';
	
	expire.setTime( +expire - 1000 );
	expire = expire.toUTCString();
	// set cookie: uname
	this.server.SetCookie( r, { name:'uname', val:'', expires:expire, domain:domain, path:'/' }, false );
	this.server.SetCookie( r, { name:'sid', val:'', expires:expire, domain:domain, path:'/' }, true );
	// set redirect
	r.status = 301;
	this.server.SetHeader( r, 'Location', '/', false );
	this.DeadEnd( true, r );
};

